/*
  pinctrl.cpp - GPIO Control Library for Linux GPIO Character Device

  Copyright (c) 2022 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <poll.h>
#include <pthread.h>
#include <sys/ioctl.h>
#include <sys/epoll.h>
#include "pinctrl.h"

#define GPIO_DEVICE "/dev/gpiochip%d"

PinCtrl GPIO;

PinCtrl::PinCtrl(void)
: _chip_count(0)
{
  memset(_chip_fd, -1, sizeof(_chip_fd));
  memset(_line_fd, -1, sizeof(_line_fd));
  memset(_pipe_fd, -1, sizeof(_pipe_fd));
}

PinCtrl::~PinCtrl(void)
{
  end();
}

size_t PinCtrl::begin(void)
{
  end();
  for (_chip_count = 0; _chip_count < PINCTRL_MAX_CHIPS; ++_chip_count)
  {
    char name[64];
    int& chip = _chip_fd[_chip_count];
    struct gpiochip_info& info = _chip_info[_chip_count];
    snprintf(name, sizeof(name), GPIO_DEVICE, _chip_count);
    if ((chip = open(name, O_RDWR | O_CLOEXEC)) < 0)
    {
      if (_chip_count == 0)
        fprintf(stderr, "[%s:%u] open(\"%s\", O_RDWR | O_CLOEXEC): %s\n", __FILE__, __LINE__, name, strerror(errno));
      break;
    }
    if (ioctl(chip, GPIO_GET_CHIPINFO_IOCTL, &info) < 0)
    {
      fprintf(stderr, "[%s:%u] ioctl(%d, GPIO_GET_CHIPINFO_IOCTL, 0x%08X): %s\n", __FILE__, __LINE__, chip, &info, strerror(errno));
      close(chip);
      chip = -1;
      break;
    }
  }
  return _chip_count;
}

const struct gpiochip_info *PinCtrl::chip_info(size_t num)
{
  return num < _chip_count ? &_chip_info[num] : nullptr;
}

void PinCtrl::end(void)
{
  for (pin_size_t i = 0; i < PINCTRL_MAX_LINES; ++i)
    release(i);
  for (int i = 0; i < PINCTRL_MAX_CHIPS; ++i)
  {
    int& chip = _chip_fd[i];
    if (chip >= 0)
    {
      close(chip);
      chip = -1;
    }
  }
  _chip_count = 0;
}

void PinCtrl::release(pin_size_t pin)
{
  stop(pin);
  if (pin < PINCTRL_MAX_LINES)
  {
    int& line = _line_fd[pin];
    if (line >= 0)
    {
      close(line);
      line = -1;
    }
  }
}

void PinCtrl::stop(pin_size_t pin)
{
  int& pipe = _pipe_fd[pin];
  if (pipe >= 0)
  {
    write(pipe, "Q", 1); // terminate polling thread
    close(pipe);
    pipe = -1;
  }
}

int PinCtrl::config(pin_size_t pin, gpio_v2_line_config& cfg)
{
  if (pin < PINCTRL_MAX_LINES)
  {
    int chip = _chip_fd[pin / GPIO_V2_LINES_MAX];
    if (chip >= 0)
    {
      int& line = _line_fd[pin];
      if (line >= 0)
      {
        stop(pin);
        if (ioctl(line, GPIO_V2_LINE_SET_CONFIG_IOCTL, &cfg) < 0)
          fprintf(stderr, "[%s:%u] ioctl(%d, GPIO_V2_LINE_SET_CONFIG_IOCTL, 0x%08X): %s\n", __FILE__, __LINE__, chip, &cfg, strerror(errno));
        else
          return line;
      }
      else
      {
        struct gpio_v2_line_request req = {.offsets = {(uint32_t)(pin % GPIO_V2_LINES_MAX)}, .config = cfg, .num_lines = 1};
        if (ioctl(chip, GPIO_V2_GET_LINE_IOCTL, &req) < 0)
          fprintf(stderr, "[%s:%u] ioctl(%d, GPIO_V2_GET_LINE_IOCTL, 0x%08X): %s\n", __FILE__, __LINE__, chip, &req, strerror(errno));
        else
          return line = req.fd;
      }
    }
  }
  return -1;
}

void PinCtrl::pinMode(pin_size_t pin, PinMode mode, PinPull pull, uint32_t debounce_period_us)
{
  struct gpio_v2_line_config cfg = {.num_attrs = (mode == INPUT), .attrs =
    {{.attr = {.id = GPIO_V2_LINE_ATTR_ID_DEBOUNCE, .debounce_period_us = debounce_period_us}, .mask = 1}}};
  switch (mode)
  {
    default:
    case INPUT:       cfg.flags = GPIO_V2_LINE_FLAG_INPUT;  break;
    case OUTPUT:      cfg.flags = GPIO_V2_LINE_FLAG_OUTPUT; break;
    case OPEN_DRAIN:  cfg.flags = GPIO_V2_LINE_FLAG_OUTPUT | GPIO_V2_LINE_FLAG_OPEN_DRAIN;  break;
    case OPEN_SOURCE: cfg.flags = GPIO_V2_LINE_FLAG_OUTPUT | GPIO_V2_LINE_FLAG_OPEN_SOURCE; break;
  }
  switch (pull)
  {
    default:
    case PULL_DISABLE: cfg.flags |= GPIO_V2_LINE_FLAG_BIAS_DISABLED;  break;
    case PULL_UP:      cfg.flags |= GPIO_V2_LINE_FLAG_BIAS_PULL_UP;   break;
    case PULL_DOWN:    cfg.flags |= GPIO_V2_LINE_FLAG_BIAS_PULL_DOWN; break;
  }
  config(pin, cfg);
}

void PinCtrl::digitalWrite(pin_size_t pin, PinStatus value)
{
  if (pin < PINCTRL_MAX_LINES)
  {
    int line = _line_fd[pin];
    if (line >= 0)
    {
      struct gpio_v2_line_values val = {.bits = (value != 0), .mask = 1};
      if (ioctl(line, GPIO_V2_LINE_SET_VALUES_IOCTL, &val) < 0)
        fprintf(stderr, "[%s:%u] ioctl(%d, GPIO_V2_LINE_SET_VALUES_IOCTL, 0x%08X): %s\n", __FILE__, __LINE__, line, &val, strerror(errno));
    }
  }
}

PinStatus PinCtrl::digitalRead(pin_size_t pin)
{
  if (pin < PINCTRL_MAX_LINES)
  {
    int line = _line_fd[pin];
    if (line >= 0)
    {
      struct gpio_v2_line_values val = {.bits = 0, .mask = 1};
      if (ioctl(line, GPIO_V2_LINE_GET_VALUES_IOCTL, &val) < 0)
        fprintf(stderr, "[%s:%u] ioctl(%d, GPIO_V2_LINE_GET_VALUES_IOCTL, 0x%08X): %s\n", __FILE__, __LINE__, line, &val, strerror(errno));
      else
        return val.bits & 1 ? HIGH : LOW;
    }
  }
  return LOW;
}

typedef struct
{
  pinctrl_callback_t func;
  void *arg;
  int fd;
  int pipe;
} polling_params_t;

static void *polling(void *arg)
{
  polling_params_t params = *(polling_params_t *)arg;
  free(arg);

  int epoll;
  if ((epoll = epoll_create1(EPOLL_CLOEXEC)) < 0)
    fprintf(stderr, "[%s:%u] epoll_create1(EPOLL_CLOEXEC): %s\n", __FILE__, __LINE__, strerror(errno));
  else
  {
    struct epoll_event ev[2];
    ev[0].events  = EPOLLIN;
    ev[0].data.fd = params.fd;
    ev[1].events  = EPOLLIN;
    ev[1].data.fd = params.pipe;
    if (epoll_ctl(epoll, EPOLL_CTL_ADD, ev[0].data.fd, &ev[0]) < 0)
      fprintf(stderr, "[%s:%u] epoll_ctl(%d, EPOLL_CTL_ADD, %d, 0x%08X): %s\n", __FILE__, __LINE__, epoll, ev[0].data.fd, &ev[0], strerror(errno));
    else if (epoll_ctl(epoll, EPOLL_CTL_ADD, ev[1].data.fd, &ev[1]) < 0)
      fprintf(stderr, "[%s:%u] epoll_ctl(%d, EPOLL_CTL_ADD, %d, 0x%08X): %s\n", __FILE__, __LINE__, epoll, ev[1].data.fd, &ev[1], strerror(errno));
    else
    {
      struct epoll_event events[PINCTRL_MAX_EVENTS];
      for (int nfds = 0; nfds >= 0; )
      {
        nfds = epoll_wait(epoll, events, PINCTRL_MAX_EVENTS, -1);
        if (nfds < 0)
          fprintf(stderr, "[%s:%u] epoll_wait(%d, 0x%08X, %d, -1): %s\n", __FILE__, __LINE__, epoll, events, PINCTRL_MAX_EVENTS, strerror(errno));
        else
        {
          for (int i = 0; i < nfds; ++i)
          {
            struct gpio_v2_line_event evt;
            int fd = events[i].data.fd;
            if (fd == params.pipe)
            {
              // fprintf(stderr, "[%s:%u] terminate polling thread\n", __FILE__, __LINE__);
              nfds = -1;
              break;
            }
            ssize_t n = read(fd, &evt, sizeof(evt));
            if (n != sizeof(evt))
            {
              fprintf(stderr, "[%s:%u] read(%d, 0x%08X, %d) != %d: %s\n", __FILE__, __LINE__, fd, &evt, sizeof(evt), n, "error reading line event");
              nfds = -1;
              break;
            }
            params.func(params.arg, evt);
          }
        }
      }
    }
    close(epoll);
  }
  close(params.pipe);
  return NULL;
}

#define GPIO_V2_LINE_FLAG_EDGE_BOTH (GPIO_V2_LINE_FLAG_EDGE_RISING | GPIO_V2_LINE_FLAG_EDGE_FALLING)

void PinCtrl::attachInterruptParam(pin_size_t pin, pinctrl_callback_t func, void *arg, PinEdge edge, PinPull pull, uint32_t debounce_period_us)
{
  struct gpio_v2_line_config cfg = {.num_attrs = 1, .attrs =
    {{.attr = {.id = GPIO_V2_LINE_ATTR_ID_DEBOUNCE, .debounce_period_us = debounce_period_us}, .mask = 1}}};
  switch (edge)
  {
    default:
    case CHANGE:  cfg.flags = GPIO_V2_LINE_FLAG_INPUT | GPIO_V2_LINE_FLAG_EDGE_BOTH;    break;
    case RISING:  cfg.flags = GPIO_V2_LINE_FLAG_INPUT | GPIO_V2_LINE_FLAG_EDGE_RISING;  break;
    case FALLING: cfg.flags = GPIO_V2_LINE_FLAG_INPUT | GPIO_V2_LINE_FLAG_EDGE_FALLING; break;
  }
  switch (pull)
  {
    default:
    case PULL_DISABLE: cfg.flags |= GPIO_V2_LINE_FLAG_BIAS_DISABLED;  break;
    case PULL_UP:      cfg.flags |= GPIO_V2_LINE_FLAG_BIAS_PULL_UP;   break;
    case PULL_DOWN:    cfg.flags |= GPIO_V2_LINE_FLAG_BIAS_PULL_DOWN; break;
  }
  int fd;
  if ((fd = config(pin, cfg)) >= 0)
  {
    int filedes[2];
    if (pipe(filedes) < 0)
      fprintf(stderr, "[%s:%u] pipe(0x%08X): %s\n", __FILE__, __LINE__, filedes, strerror(errno));
    else
    {
      polling_params_t *params = (polling_params_t *)malloc(sizeof(*params));
      if (!params)
        fprintf(stderr, "[%s:%u] malloc(%u): %s\n", __FILE__, __LINE__, (uint32_t)sizeof(*params), "insufficient memory");
      else
      {
        pthread_t thid;
        params->func = func;
        params->arg  = arg;
        params->fd   = fd;
        params->pipe = filedes[0];
        if (pthread_create(&thid, NULL, polling, params) < 0)
          fprintf(stderr, "[%s:%u] pthread_create(0x%08X, NULL, 0x%08X, 0x%08X): %s\n", __FILE__, __LINE__, &thid, polling, params, strerror(errno));
        else
        {
          _pipe_fd[pin] = filedes[1];
          return; // success
        }
        free(params);
      }
      close(filedes[0]);
      close(filedes[1]);
    }
  }
}

void PinCtrl::detachInterrupt(pin_size_t pin)
{
  release(pin);
}

void PinCtrl::sleepMicroseconds(size_t usec)
{
  struct timespec rem, req = { tv_sec: (time_t)(usec / 1000000), tv_nsec: (long)((usec % 1000000) * 1000) };
  while (nanosleep(&req, &rem) == EINTR)
    req = rem;
}

void PinCtrl::sleep(size_t msec)
{
  struct timespec rem, req = { tv_sec: (time_t)(msec / 1000), tv_nsec: (long)((msec % 1000) * 1000000) };
  while (nanosleep(&req, &rem) == EINTR)
    req = rem;
}

void PinCtrl::delayMicroseconds(size_t usec)
{
  size_t t = micros();
  while (micros() - t < usec)
    continue;
}

void PinCtrl::delay(size_t msec)
{
  size_t t = millis();
  while (millis() - t < msec)
    continue;
}

size_t PinCtrl::micros(void)
{
  struct timespec ts;

  clock_gettime(CLOCK_MONOTONIC_RAW, &ts);
  return (size_t)ts.tv_sec * 1000000 + (ts.tv_nsec / 1000);
}

size_t PinCtrl::millis(void)
{
  struct timespec ts;

  clock_gettime(CLOCK_MONOTONIC_RAW, &ts);
  return (size_t)ts.tv_sec * 1000 + (ts.tv_nsec / 1000000);
}
